import pytest
import os
import shutil

from kestrel.session import Session
from kestrel.exceptions import MissingEntityType


def test_load_full_csv():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs.csv"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_load_relative_path_csv(tmp_path):
    data_file_path = "test_input_data_procs.csv"
    ori_path = os.path.join(
        os.path.dirname(__file__), data_file_path
    )
    shutil.copy2(ori_path, tmp_path)
    os.chdir(tmp_path)
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_load_full_json():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs.json"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_load_parquet_gz():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs.parquet.gz"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_load_notype_json_to_fail():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs_no_type.json"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        with pytest.raises(MissingEntityType) as e:
            s.execute(stmt)


def test_load_notype_json_as_type():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs_no_type.json"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path} AS process"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_load_string_list_to_fail():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs_list.json"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path}"
        with pytest.raises(MissingEntityType) as e:
            s.execute(stmt)


def test_load_string_list_as_type():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs_list.json"
    )
    with Session() as s:
        stmt = f"newvar = LOAD {data_file_path} AS process"
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 3
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "reg.exe", "explorer.exe"]
        assert sorted([i["name"] for i in v]) == ["cmd.exe", "explorer.exe", "reg.exe"]



################################################################
#             Batch Execution of Kestrel Huntflow
#
# Usage: `python3 -m kestrel [-v] [--debug] hunting.hf`
################################################################

import argparse
import logging

from kestrel.session import Session
from kestrel.utils import add_logging_handler, clear_logging_handlers


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Kestrel Interpreter")
    parser.add_argument("huntflow", help="huntflow in .hf file")
    parser.add_argument(
        "-v", "--verbose", help="print verbose log", action="store_true"
    )
    parser.add_argument(
        "--debug", help="debug level log (default is info level)", action="store_true"
    )
    args = parser.parse_args()

    clear_logging_handlers()
    if args.verbose:
        add_logging_handler(logging.StreamHandler(), args.debug)

    with Session(debug_mode=args.debug) as session:
        with open(args.huntflow, "r") as fp:
            huntflow = fp.read()
        outputs = session.execute(huntflow)
        results = "\n\n".join([o.to_string() for o in outputs])
        print(results)



import logging
from copy import deepcopy
from kestrel.symboltable.symtable import SymbolTable
from kestrel.syntax.parser import parse_ecgpattern
from kestrel.utils import lowered_str_list
from kestrel.semantics.reference import make_deref_func, make_var_timerange_func
from kestrel.syntax.utils import (
    timedelta_seconds,
)
from kestrel.codegen.relations import (
    get_entity_id_attribute,
    fine_grained_relational_process_filtering,
    compile_identical_entity_search_pattern,
    build_pattern_from_ids,
)


_logger = logging.getLogger(__name__)


def do_prefetch(
    local_stage_varname,
    local_stage_varstruct,
    session,
    stmt,
    extra_prefetch_enabled_flag=True,
):
    """prefetch identical entities and associated entities.

    Put the input entities in the center of an observation and query the remote
    data source of associated with input variable, so we get back:

    1. all records about the input entities.

    2. associated entities such as parent/child processes of processes, processes of network-traffic, etc.

    Args:
        local_stage_varname (str): local variable name before prefetch
        local_stage_varstruct (str): local variable (VarStruct) before prefetch
    Returns:
        str: the entity table for the statement output variable
    """

    # STIX pattern for prefetch
    stix_pattern = None

    prefetch_ret_entity_table = stmt["output"] + "_prefetch"
    prefetch_fil_entity_table = prefetch_ret_entity_table + "_filtered"
    return_var_entity_table = stmt["output"]

    if (
        _is_prefetch_allowed_in_config(
            session.config["prefetch"], stmt["command"], stmt["type"]
        )
        and len(local_stage_varstruct)
        and local_stage_varstruct.data_source
        and extra_prefetch_enabled_flag
    ):
        _logger.info(f"generate pattern for prefetch {local_stage_varname}.")

        pattern_raw = compile_identical_entity_search_pattern(
            local_stage_varname,
            local_stage_varstruct,
            session.config["stixquery"]["support_id"],
        )

        if pattern_raw:
            _symtable = SymbolTable({local_stage_varname: local_stage_varstruct})
            pattern_ast = parse_ecgpattern(pattern_raw)
            deref_func = make_deref_func(session.store, _symtable)
            get_timerange_func = make_var_timerange_func(session.store, _symtable)
            pattern_ast.deref(deref_func, get_timerange_func)
            pattern_ast.add_center_entity(local_stage_varstruct.type)
            time_adj = tuple(
                map(
                    timedelta_seconds,
                    (
                        session.config["stixquery"]["timerange_start_offset"],
                        session.config["stixquery"]["timerange_stop_offset"],
                    ),
                )
            )
            ext_graph_pattern = deepcopy(stmt.get("where"))
            # TODO: check prune() logic, remove the "if" and just run content
            if ext_graph_pattern and stmt["command"].lower() == "get":
                ext_graph_pattern.prune_away_centered_graph(stmt["type"])
            _logger.debug(f"ext pattern in prefetch: {ext_graph_pattern}")
            _logger.debug(f"prefetch pattern before extend: {pattern_ast}")
            pattern_ast.extend("AND", ext_graph_pattern)
            _logger.debug(f"prefetch pattern after extend: {pattern_ast}")
            stix_pattern = pattern_ast.to_stix(stmt["timerange"], time_adj)
            _logger.info(f"STIX pattern generated in prefetch: {stix_pattern}")

    if stix_pattern:
        resp = session.data_source_manager.query(
            local_stage_varstruct.data_source,
            stix_pattern,
            session.session_id,
            session.store,
            stmt.get("limit"),
        )
        query_id = resp.load_to_store(session.store)

        # build the view in store
        session.store.extract(
            prefetch_ret_entity_table, stmt["type"], query_id, stix_pattern
        )

        prefetch_final_entity_table = _filter_prefetched_process(
            prefetch_fil_entity_table,
            session,
            local_stage_varstruct,
            prefetch_ret_entity_table,
            stmt["type"],
        )

        session.store.rename_view(prefetch_final_entity_table, return_var_entity_table)

        for v in [
            local_stage_varstruct.entity_table,
            prefetch_ret_entity_table,
            prefetch_fil_entity_table,
        ]:
            if not session.debug_mode:
                session.store.remove_view(v)

    else:
        _logger.info(f"prefetch does not happen without STIX pattern generated.")
        session.store.rename_view(
            local_stage_varstruct.entity_table, return_var_entity_table
        )

    return return_var_entity_table


def _filter_prefetched_process(
    prefetch_filtered_var_name,
    session,
    local_varstruct,
    prefetch_raw_entity_table,
    return_type,
):
    # default return
    ret_table = prefetch_raw_entity_table

    # special handling for process to filter out impossible relational processes
    # this is needed since STIX 2.0 does not have mandatory fields for
    # process and field like `pid` is not unique
    if return_type == "process" and get_entity_id_attribute(local_varstruct) not in (
        "id",
        "x_unique_id",
    ):
        _logger.debug(
            f"filter prefetched {return_type} for {prefetch_raw_entity_table}."
        )

        entity_ids = fine_grained_relational_process_filtering(
            local_varstruct,
            prefetch_raw_entity_table,
            session.store,
            session.config["prefetch"]["process_identification"],
        )

        if entity_ids:
            id_pattern = build_pattern_from_ids(return_type, entity_ids)
            session.store.extract(
                prefetch_filtered_var_name, return_type, None, id_pattern
            )
            _logger.debug("filter after prefetch succeeds.")
            ret_table = prefetch_filtered_var_name
        else:
            _logger.info(
                "no prefetched process found after filtering, return raw prefetch."
            )

    return ret_table


def _is_prefetch_allowed_in_config(prefetch_config, command_name, entity_type):
    if prefetch_config["switch_per_command"][
        command_name
    ] and entity_type not in lowered_str_list(prefetch_config["excluded_entities"]):
        return True
    else:
        return False



import logging
import asyncio
import copy
from typing import Union
from typeguard import typechecked
from pandas import DataFrame
from multiprocessing import Queue

from kestrel.datasource import ReturnFromStore
from kestrel.utils import mkdtemp
from kestrel.exceptions import DataSourceError, DataSourceManagerInternalError
from kestrel_datasource_stixshifter.connector import check_module_availability
from kestrel_datasource_stixshifter import multiproc
from kestrel_datasource_stixshifter.config import (
    get_datasource_from_profiles,
    load_options,
    load_profiles,
    set_stixshifter_logging_level,
)

from firepit.sqlstorage import SqlStorage
import firepit.aio.ingest
import firepit.aio.asyncwrapper

from stix_shifter.stix_translation import stix_translation

# TODO: better solution to avoid using nest_asyncio for run_until_complete()
#       maybe putting entire Kestrel in async mode
import nest_asyncio

nest_asyncio.apply()


_logger = logging.getLogger(__name__)


def query_datasource(uri, pattern, session_id, config, store, limit=None):
    # CONFIG command is not supported
    # profiles will be updated according to YAML file and env var
    config["profiles"] = load_profiles()

    config["options"] = load_options()

    _logger.debug(
        "fast_translate enabled for: " f"{config['options']['fast_translate']}"
    )

    scheme, _, profile = uri.rpartition("://")
    profiles = profile.split(",")
    if scheme != "stixshifter":
        raise DataSourceManagerInternalError(
            f"interface {__package__} should not process scheme {scheme}"
        )

    set_stixshifter_logging_level()

    # cache_dir will be empty if not in debug mode
    # however, the directory guarantees unique and non-exist query ID
    # so it always needs to be created
    cache_dir = mkdtemp()
    query_id = cache_dir.name

    _logger.debug(f"prepare query with ID: {query_id}")

    num_records = 0
    profile_limit = limit

    for profile in profiles:
        if limit:
            if num_records >= limit:
                break
            if num_records > 0:
                profile_limit = limit - num_records
        _logger.debug(f"entering stix-shifter data source: {profile}")
        _logger.debug(f"profile = {profile}, profile_limit = {profile_limit}")
        # STIX-shifter will alter the config objects, thus making them not reusable.
        # So only give STIX-shifter a copy of the configs.
        # Check `modernize` functions in the `stix_shifter_utils` for details.
        (
            connector_name,
            connection_dict,
            configuration_dict,
            retrieval_batch_size,
            cool_down_after_transmission,
        ) = map(
            copy.deepcopy, get_datasource_from_profiles(profile, config["profiles"])
        )

        check_module_availability(connector_name)

        if _logger.isEnabledFor(logging.DEBUG):
            data_path_striped = "".join(filter(str.isalnum, profile))
            cache_data_path_prefix = str(cache_dir / data_path_striped)
        else:
            cache_data_path_prefix = None

        observation_metadata = gen_observation_metadata(connector_name, query_id)

        dsl = translate_query(
            connector_name, observation_metadata, pattern, connection_dict
        )

        raw_records_queue = Queue()
        translated_data_queue = Queue()

        exceptions = []

        with multiproc.translate(
            connector_name,
            observation_metadata,
            connection_dict.get("options", {}),
            cache_data_path_prefix,
            connector_name in config["options"]["fast_translate"],
            raw_records_queue,
            translated_data_queue,
            config["options"]["translation_workers_count"],
        ):
            with multiproc.transmit(
                connector_name,
                connection_dict,
                configuration_dict,
                retrieval_batch_size,
                config["options"]["translation_workers_count"],
                cool_down_after_transmission,
                dsl["queries"],
                raw_records_queue,
                profile_limit,
            ):
                for result in multiproc.read_translated_results(
                    translated_data_queue,
                    config["options"]["translation_workers_count"],
                ):
                    num_records += get_num_objects(result)
                    ingest(result, observation_metadata, query_id, store)

    return ReturnFromStore(query_id)


@typechecked
def gen_observation_metadata(connector_name: str, query_id: str):
    return {
        "id": "identity--" + query_id,
        "name": connector_name,
        "type": "identity",
    }


@typechecked
def translate_query(
    connector_name: str,
    observation_metadata: dict,
    pattern: str,
    connection_dict: dict,
):
    translation = stix_translation.StixTranslation()
    translation_options = copy.deepcopy(connection_dict.get("options", {}))

    _logger.debug(f"STIX pattern to query: {pattern}")

    dsl = translation.translate(
        connector_name, "query", observation_metadata, pattern, translation_options
    )

    _logger.debug(f"translate results: {dsl}")

    if "error" in dsl:
        raise DataSourceError(
            "STIX-shifter translation from STIX"
            " to native query failed"
            f" with message: {dsl['error']}"
        )

    return dsl


@typechecked
def ingest(
    result: Union[dict, DataFrame],
    observation_metadata: dict,
    query_id: str,
    store: SqlStorage,
):
    _logger.debug("ingestion of a batch/page starts")
    if isinstance(result, DataFrame):
        # fast translation result in DataFrame
        asyncio.run(
            firepit.aio.ingest.ingest(
                firepit.aio.asyncwrapper.SyncWrapper(store=store),
                observation_metadata,
                result,
                query_id,
            )
        )
    else:
        # STIX bundle (normal stix-shifter translation result)
        store.cache(query_id, result)
    _logger.debug("ingestion of a batch/page ends")


@typechecked
def get_num_objects(data: Union[dict, DataFrame]):
    if isinstance(data, DataFrame):
        num_objects = len(data)
    else:
        num_objects = len(data.get("objects", []))
        if num_objects > 0:
            num_objects -= 1  # minus the identify object
    return num_objects



from typing import Optional, Union, List
from dataclasses import dataclass
from pandas import DataFrame

STOP_SIGN = "STOP"


@dataclass
class WorkerLog:
    level: int
    log: str


# if success == True, data and offset is not None
# if success == False, log is not None
@dataclass
class TransmissionResult:
    worker: str
    success: bool
    data: Optional[List[dict]]
    offset: Optional[int]
    log: Optional[WorkerLog]


# if success == True, data is not None
# if success == False, log is not None
@dataclass
class TranslationResult:
    worker: str
    success: bool
    data: Union[None, dict, DataFrame]
    log: Optional[WorkerLog]



import os
import pytest

from kestrel.session import Session


NEW_PROCS = """
p = NEW [
          {"type": "process", "name": "cmd.exe", "command_line": "cmd -c dir"},
          {"type": "process", "name": "explorer.exe", "pid": "99"}
        ]
"""

REF_PROCS = """
ref = NEW [
          {"type": "process", "name": "", "pid": 4},
          {"type": "process", "name": "explorer.exe", "pid": 1380}
        ]
"""

@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "doctored-1k.json")


@pytest.mark.parametrize(
    "stmt, expected",
    [
        ("x = p", 2),
        ("x = p WHERE pid = 99", 1),
        ("x = p WHERE command_line IS NULL", 1),
        ("x = p WHERE command_line IS NOT NULL", 1),
        ("x = p WHERE command_line LIKE '%cmd%'", 1),
    ],
)
def test_assign_after_new(stmt, expected):
    with Session() as s:
        s.execute(NEW_PROCS)
        s.execute(stmt)
        x = s.get_variable("x")
        assert len(x) == expected, f"ASSIGN error: f{stmt}"


# The * 2 on these counts is due to our inability to dedup process objects
# Need unique IDs on process objects
@pytest.mark.parametrize(
    "stmt, expected",
    [
        ("x = p", 1000),
        ("x = p WHERE pid = 1380", 106),
        ("x = p WHERE command_line IS NULL", 948),
        ("x = p WHERE command_line IS NOT NULL", 52),
        ("x = p WHERE command_line LIKE '%/node%'", 1),
        ("x = p WHERE pid = 5960 OR name = 'taskeng.exe'", 2),
        ("x = p WHERE (pid = 5960 OR name = 'taskeng.exe') AND command_line IS NULL", 0),
    ],
)
def test_assign_after_get(proc_bundle_file, stmt, expected):
    with Session() as s:
        s.execute(f"""
                   p = GET process
                       FROM file://{proc_bundle_file}
                       WHERE [process:pid > 0]
                   """
        )
        s.execute(stmt)
        x = s.get_variable("x")
        assert len(x) == expected, f"ASSIGN error: {stmt}"


def test_assign_with_reference(proc_bundle_file):
    with Session() as s:
        s.execute(f"p = GET process FROM file://{proc_bundle_file} WHERE [process:pid > 0]")
        s.execute(REF_PROCS)
        s.execute("q = p WHERE pid = ref.pid")
        q = s.get_variable("q")
        assert len(q) == 106 + 149


def test_assign_with_reference_and_in(proc_bundle_file):
    with Session() as s:
        s.execute(f"p = GET process FROM file://{proc_bundle_file} WHERE [process:pid > 0]")
        p = s.get_variable("p", False)
        assert 'binary_ref' in p[0]
        s.execute(REF_PROCS)
        s.execute("q = p WHERE pid IN (ref.pid, 9240, 10020)")
        q = s.get_variable("q", False)
        assert len(q) == 106 + 149 + 1 + 1
        print(q[0])
        assert 'binary_ref' in q[0]


def test_assign_from_empty_var(proc_bundle_file):
    with Session() as s:
        s.execute(f"p = GET process FROM file://{proc_bundle_file} WHERE [process:pid < 0]")
        s.execute(REF_PROCS)
        s.execute("q = p WHERE pid = ref.pid")
        q = s.symtable["q"]
        assert len(q) == 0
        assert q.records_count == 0



import json
import os
import shutil

import pytest

from kestrel.codegen.display import DisplayWarning
from kestrel.session import Session

from .utils import set_empty_kestrel_config, set_no_prefetch_kestrel_config


@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "doctored-1k.json")


@pytest.fixture()
def file_stix_bundles():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return [
        os.path.join(cwd, "test_bundle_4.json"),
        os.path.join(cwd, "test_bundle_5.json"),
    ]


@pytest.fixture()
def nt_stix_bundles():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return [
        os.path.join(cwd, "test_bundle_nt_1.json"),
        os.path.join(cwd, "test_bundle_nt_2.json"),
    ]


@pytest.fixture()
def set_stixshifter_stix_bundles():
    cfg = '{"auth": {"username": "","password": ""}}'
    connector = "stix_bundle"
    stixshifter_data_url = "https://raw.githubusercontent.com/opencybersecurityalliance/stix-shifter/develop/data/cybox"
    host1 = f"{stixshifter_data_url}/carbon_black/cb_observed_156.json"
    host2 = f"{stixshifter_data_url}/qradar/qradar_custom_process_observable.json"

    os.environ["STIXSHIFTER_HOST1_CONNECTION"] = json.dumps({"host": host1})
    os.environ["STIXSHIFTER_HOST1_CONNECTOR"] = connector
    os.environ["STIXSHIFTER_HOST1_CONFIG"] = cfg
    os.environ["STIXSHIFTER_HOST2_CONNECTION"] = json.dumps({"host": host2})
    os.environ["STIXSHIFTER_HOST2_CONNECTOR"] = connector
    os.environ["STIXSHIFTER_HOST2_CONFIG"] = cfg

    # use `yield` to pause here before a test finished
    # https://docs.pytest.org/en/latest/how-to/fixtures.html#teardown-cleanup-aka-fixture-finalization
    yield None

    # this clean up will be executed when the test (that uses the fixture) exits
    ss_envs = [k for k in list(os.environ.keys()) if k.startswith("STIXSHIFTER_")]
    for ss_env in ss_envs:
        del os.environ[ss_env]


def test_get_single_file(file_stix_bundles):
    with Session() as s:
        stmt = f"""
                var = GET process
                      FROM file://{file_stix_bundles[0]}
                      WHERE [process:name='compattelrunner.exe']
                """

        s.execute(stmt)
        v = s.get_variable("var")
        print(json.dumps(v, indent=4))
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_single_file_limit(file_stix_bundles):
    with Session() as s:
        stmt = f"""
                var = GET process
                      FROM file://{file_stix_bundles[0]}
                      WHERE [process:name='compattelrunner.exe']
                      LIMIT 1
                """

        s.execute(stmt)
        v = s.get_variable("var")
        print(json.dumps(v, indent=4))
        assert len(v) == 1
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_single_file_limit_1(file_stix_bundles):
    with Session() as s:
        stmt = f"""
                var = GET process
                      FROM file://{file_stix_bundles[0]}
                      WHERE [process:name='compattelrunner.exe']
                      LIMIT 10
                """

        s.execute(stmt)
        v = s.get_variable("var")
        print(json.dumps(v, indent=4))
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_multiple_file_stix_bundles(file_stix_bundles):
    with Session() as s:
        file_bundles = ",".join(file_stix_bundles)
        stmt = f"""
                var = GET process
                      FROM file://{file_bundles}
                      WHERE name = 'compattelrunner.exe'
                """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_multiple_file_stix_bundles_limit(file_stix_bundles):
    with Session() as s:
        file_bundles = ",".join(file_stix_bundles)
        stmt = f"""
                var = GET process
                      FROM file://{file_bundles}
                      WHERE name = 'compattelrunner.exe'
                      LIMIT 3
                """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 3
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_multiple_file_stix_bundles_limit_1(file_stix_bundles):
    with Session() as s:
        file_bundles = ",".join(file_stix_bundles)
        stmt = f"""
                var = GET process
                      FROM file://{file_bundles}
                      WHERE name = 'compattelrunner.exe'
                      LIMIT 4
                """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 4
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


def test_get_multiple_file_stix_bundles_limit_2(file_stix_bundles):
    with Session() as s:
        file_bundles = ",".join(file_stix_bundles)
        stmt = f"""
                var = GET process
                      FROM file://{file_bundles}
                      WHERE name = 'compattelrunner.exe'
                      LIMIT 8
                """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "compattelrunner.exe"


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_single_stixshifter_stix_bundle(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST2
                     WHERE [ipv4-addr:value = '127.0.0.1']
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
               """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 6
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] == "powershell.exe"


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_single_stixshifter_stix_bundle_limit(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST2
                     WHERE [ipv4-addr:value = '127.0.0.1']
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
                     LIMIT 4
               """

        s.execute(stmt)
        v = s.get_variable("var")
        assert len(v) == 4
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] == "powershell.exe"


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_multiple_stixshifter_stix_bundles(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST1,HOST2
                     WHERE ipv4-addr:value = '127.0.0.1'
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
               """

        s.execute(stmt)
        v = s.get_variable("var")

        # The extended graph [ipv4-addr:value = '127.0.0.1'] is recognized and
        # merged to prefetch query, resultsing in limited (32) processes. If
        # not used by prefetch, the total number of process records prefetched
        # is 240.
        assert len(v) == 32
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] in [
                "powershell.exe",
                "(unknown)",
                "explorer.exe",
                "firefox.exe",
                "ntoskrnl.exe",
                "teamviewer_service.exe",
                "teamviewer.exe",
                "vmware.exe",
                "dashost.exe",
                "applemobiledeviceservice.exe",
                "svctest.exe",
                "vmware-hostd.exe",
            ]


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_multiple_stixshifter_stix_bundles_limit(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST1,HOST2
                     WHERE ipv4-addr:value = '127.0.0.1'
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
                     LIMIT 10
               """

        s.execute(stmt)
        v = s.get_variable("var")

        # The extended graph [ipv4-addr:value = '127.0.0.1'] is recognized and
        # merged to prefetch query, resultsing in limited (32) processes. If
        # not used by prefetch, the total number of process records prefetched
        # is 240.
        assert len(v) == 20
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] in [
                "powershell.exe",
                "(unknown)",
                "explorer.exe",
                "firefox.exe",
                "ntoskrnl.exe",
                "teamviewer_service.exe",
                "teamviewer.exe",
                "vmware.exe",
                "dashost.exe",
                "applemobiledeviceservice.exe",
                "svctest.exe",
                "vmware-hostd.exe",
            ]


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_multiple_stixshifter_stix_bundles_limit_1(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST1,HOST2
                     WHERE ipv4-addr:value = '127.0.0.1'
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
                     LIMIT 15
               """

        s.execute(stmt)
        v = s.get_variable("var")

        # The extended graph [ipv4-addr:value = '127.0.0.1'] is recognized and
        # merged to prefetch query, resultsing in limited (32) processes. If
        # not used by prefetch, the total number of process records prefetched
        # is 240.
        assert len(v) == 28
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] in [
                "powershell.exe",
                "(unknown)",
                "explorer.exe",
                "firefox.exe",
                "ntoskrnl.exe",
                "teamviewer_service.exe",
                "teamviewer.exe",
                "vmware.exe",
                "dashost.exe",
                "applemobiledeviceservice.exe",
                "svctest.exe",
                "vmware-hostd.exe",
            ]


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_get_multiple_stixshifter_stix_bundles_limit_2(set_no_prefetch_kestrel_config, set_stixshifter_stix_bundles):
    with Session() as s:
        # default data source schema is stixshifter
        stmt = """
               var = GET process
                     FROM HOST1,HOST2
                     WHERE ipv4-addr:value = '127.0.0.1'
                     START 2019-01-01T00:00:00Z STOP 2023-01-01T00:00:00Z
                     LIMIT 50
               """

        s.execute(stmt)
        v = s.get_variable("var")

        # The extended graph [ipv4-addr:value = '127.0.0.1'] is recognized and
        # merged to prefetch query, resultsing in limited (32) processes. If
        # not used by prefetch, the total number of process records prefetched
        # is 240.
        assert len(v) == 32
        for i in range(len(v)):
            assert v[i]["type"] == "process"
            assert v[i]["name"] in [
                "powershell.exe",
                "(unknown)",
                "explorer.exe",
                "firefox.exe",
                "ntoskrnl.exe",
                "teamviewer_service.exe",
                "teamviewer.exe",
                "vmware.exe",
                "dashost.exe",
                "applemobiledeviceservice.exe",
                "svctest.exe",
                "vmware-hostd.exe",
            ]


def test_last_datasource(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                a = GET process
                    FROM file://{proc_bundle_file}
                    WHERE name = "cmd.exe"
                b = GET process
                    WHERE name = 'svchost.exe'
                """

        output = s.execute(stmt)
        a = s.get_variable("a")
        b = s.get_variable("b")
        assert len(a) == 14
        assert len(b) == 704


def test_relative_file_path(tmp_path):
    data_file_path = "doctored-1k.json"
    ori_path = os.path.join(
        os.path.dirname(__file__), data_file_path
    )
    shutil.copy2(ori_path, tmp_path)
    os.chdir(tmp_path)

    with Session() as s:
        stmt = f"""
                a = GET process
                    FROM file://{data_file_path}
                    WHERE name = "cmd.exe"
                b = GET process
                    FROM file://./{data_file_path}
                    WHERE name = 'svchost.exe'
                """
        output = s.execute(stmt)
        a = s.get_variable("a")
        b = s.get_variable("b")
        assert len(a) == 14
        assert len(b) == 704


def test_get_wrong_type(file_stix_bundles):
    with Session() as s:
        stmt = f"""
                var = GET foo
                      FROM file://{file_stix_bundles[0]}
                      WHERE name = "compattelrunner.exe"
                """

        output = s.execute(stmt)
        warnings = []
        for o in output:
            print(json.dumps(o.to_dict(), indent=4))
            if isinstance(o, DisplayWarning):
                warnings.append(o)
        assert len(warnings) == 1
        assert "foo" in warnings[0].to_string()
        v = s.get_variable("var")
        print(json.dumps(v, indent=4))
        assert len(v) == 0


def test_get_repeated(proc_bundle_file):
    """process objects may not have deterministic IDs, so we need to prevent duplicate entries somehow"""
    with Session() as s:
        stmt = f"""
                var = GET process
                      FROM file://{proc_bundle_file}
                      WHERE name = "cmd.exe"
                """
        output = s.execute(stmt)
        data = output[0].to_dict()["data"]["variables updated"][0]
        print(json.dumps(data, indent=4))
        n_ent = data["#(ENTITIES)"]
        n_rec = data["#(RECORDS)"]

        # Re-run; counts should not change
        output = s.execute(stmt)
        data = output[0].to_dict()["data"]["variables updated"][0]
        assert n_ent == data["#(ENTITIES)"]
        assert n_rec == data["#(RECORDS)"]

        stmt = f"""
                var = GET file
                      FROM file://{proc_bundle_file}
                      WHERE [file:name = 'cmd.exe']
                """
        output = s.execute(stmt)
        data = output[0].to_dict()["data"]["variables updated"][0]
        print(json.dumps(data, indent=4))
        assert data["#(ENTITIES)"] > 0
        assert data["#(RECORDS)"] > 0


@pytest.mark.parametrize(
    "num, unit, count",
    [
        # They're mostly out of the time range
        (1, "d", 0),
        (10, "days", 0),
        (9, "h", 0),
        (99, "hours", 0),
        (5, "m", 0),
        (50, "minutes", 0),
        (3600, "s", 0),
        (30, "SECONDS", 0),
        (3650, "DAYS", 2),  # This will fail sometime in 2031
    ],
)
def test_get_relative_timespan(file_stix_bundles, num, unit, count):
    with Session() as s:
        stmt = f"""
                var = GET process
                      FROM file://{file_stix_bundles[0]}
                      WHERE name='compattelrunner.exe' LAST {num} {unit}
                """

        s.execute(stmt)
        v = s.get_variable("var")
        print(json.dumps(v, indent=4))
        assert len(v) == count


def test_get_referred_variable(nt_stix_bundles):
    with Session() as s:
        stmt1 = f"""
                 nt111 = GET network-traffic
                         FROM file://{nt_stix_bundles[0]}
                         WHERE dst_ref.value = '192.168.56.112'
                 """
        s.execute(stmt1)

        stmt2 = f"""
                 nt112 = GET network-traffic
                         FROM file://{nt_stix_bundles[1]}
                         WHERE src_port = nt111.src_port
                 """
        s.execute(stmt2)

        stmt3 = f"""
                 nt112y = GET network-traffic
                          FROM file://{nt_stix_bundles[1]}
                          WHERE src_ref.value = nt111.src_ref.value
                 """
        s.execute(stmt3)

        s.config["stixquery"]["timerange_start_offset"] = -10
        s.config["stixquery"]["timerange_stop_offset"] = 10
        stmt4 = f"""
                 nt112z = GET network-traffic
                          FROM file://{nt_stix_bundles[1]}
                          WHERE src_ref.value = nt111.src_ref.value
                 """
        s.execute(stmt4)

        nt112 = s.get_variable("nt112")
        nt112y = s.get_variable("nt112y")
        nt112z = s.get_variable("nt112z")

        # src_port should uniquely identify the 4 network-traffic
        # between the two hosts: 111 and 112
        assert len(nt112) == 4

        # there are many nt from 111 to 112 in a larger time range
        # by default, offset are -300, 300 seconds, which will give more results
        assert len(nt112y) == 7

        # reset the offsets to nearly 0 (need to tolerate clock sync diff)
        # now it should go back to 4
        assert len(nt112z) == 4


def test_regex_escaping_in_stix_bundle(nt_stix_bundles):
    with Session() as s:
        stmt1 = f"""
                 d = GET directory
                     FROM file://{nt_stix_bundles[0]}
                     WHERE path MATCHES 
                 """ + r"'C:\\\\Windows.*'" # FIXME: r"'C:\\Windows.*' is expected
        s.execute(stmt1)
        d = s.get_variable("d")
        assert len(d) == 1



################################################################
#                       Module Summary
#
# - Code generation for each command in kestrel.lark
#   - The execution function names match commands in kestrel.lark
# - Each command takes 2 arguments
#     ( statement, session )
#   - statement is the current statement to process,
#     which is a dict from the parser
#   - session is the current session (context)
# - Every command returns a tuple (VarStruct, Display)
#   - VarStruct is a new object associated with the output var
#     - VarStruct associated with stmt["output"]
#     - None for some commands, e.g., DISP, SAVE, STAT
#   - Display is the data to display on the user interface
#     - a string
#     - a list of (str,str|list(str)) tuples
#     - a table that can be imported to pandas dataframe
################################################################

import functools
import logging
import itertools
from collections import OrderedDict

from firepit.deref import auto_deref
from firepit.exceptions import InvalidAttr
from firepit.query import Limit, Offset, Order, Predicate, Projection, Query
from firepit.stix20 import summarize_pattern

from kestrel.utils import remove_empty_dicts, dedup_ordered_dicts
from kestrel.exceptions import *
from kestrel.symboltable.variable import new_var
from kestrel.syntax.utils import (
    get_entity_types,
    get_all_input_var_names,
)
from kestrel.codegen.data import load_data, load_data_file, dump_data_to_file
from kestrel.codegen.display import DisplayDataframe, DisplayDict, DisplayWarning
from kestrel.codegen.relations import (
    generic_relations,
    get_entity_id_attribute,
)
from kestrel.codegen.prefetch import do_prefetch
from kestrel.codegen.queries import (
    compile_specific_relation_to_query,
    compile_generic_relation_to_query,
)

_logger = logging.getLogger(__name__)


################################################################
#                       Private Decorators
################################################################


def _default_output(func):
    # by default, create a table/view in the backend
    # using the output var name
    # in this case, the store backend can return no VarStruct
    @functools.wraps(func)
    def wrapper(stmt, session):
        ret = func(stmt, session)
        if not ret:
            var_struct = new_var(
                session.store, stmt["output"], [], stmt, session.symtable
            )
            return var_struct, None
        else:
            return ret

    return wrapper


def _skip_command_if_empty_input(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        var_names = get_all_input_var_names(stmt)
        if not var_names:
            return func(stmt, session)
        elif any(
            [
                session.symtable[v].length + session.symtable[v].records_count
                for v in var_names
            ]
        ):
            return func(stmt, session)
        elif "output" in stmt:
            var_struct = new_var(session.store, None, [], stmt, session.symtable)
            return var_struct, None
        else:
            return None, None

    return wrapper


def _debug_logger(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        _logger.debug(f"Executing '{func.__name__}' with statement: {stmt}")
        return func(stmt, session)

    return wrapper


################################################################
#                 Code Generation for Commands
################################################################


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def assign(stmt, session):
    entity_table = session.symtable[stmt["input"]].entity_table
    transform = stmt.get("transformer")
    if transform:
        qry = _transform_query(session.store, entity_table, transform)
    else:
        qry = Query(entity_table)
    qry = _build_query(session.store, entity_table, qry, stmt, [])
    session.store.assign_query(stmt["output"], qry)


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def merge(stmt, session):
    entity_types = list(
        set([session.symtable[var_name].type for var_name in stmt["inputs"]])
    )
    if len(entity_types) > 1:
        raise NonUniformEntityType(entity_types)
    entity_tables = [
        session.symtable[var_name].entity_table for var_name in stmt["inputs"]
    ]
    entity_tables = [t for t in entity_tables if t is not None]
    session.store.merge(stmt["output"], entity_tables)


@_debug_logger
@_default_output
def new(stmt, session):
    stmt["type"] = load_data(session.store, stmt["output"], stmt["data"], stmt["type"])


@_debug_logger
@_default_output
def load(stmt, session):
    stmt["type"] = load_data_file(
        session.store, stmt["output"], stmt["path"], stmt["type"]
    )


@_debug_logger
def save(stmt, session):
    dump_data_to_file(
        session.store, session.symtable[stmt["input"]].entity_table, stmt["path"]
    )
    return None, None


@_debug_logger
@_skip_command_if_empty_input
def info(stmt, session):
    header = session.store.columns(session.symtable[stmt["input"]].entity_table)
    direct_attrs, associ_attrs, custom_attrs, references = [], [], [], []
    for field in header:
        if field.startswith("x_"):
            custom_attrs.append(field)
        elif (
            field.endswith("_ref")
            or field.endswith("_refs")
            or field.endswith("_reference")
            or field.endswith("_references")
        ):
            # not useful in existing version, so do not display
            references.append(field)
        elif "_ref." in field or "_ref_" in field:
            associ_attrs.append(field)
        else:
            direct_attrs.append(field)

    disp = OrderedDict()
    disp["Entity Type"] = session.symtable[stmt["input"]].type
    disp["Number of Entities"] = str(len(session.symtable[stmt["input"]]))
    disp["Number of Records"] = str(session.symtable[stmt["input"]].records_count)
    disp["Entity Attributes"] = ", ".join(direct_attrs)
    disp["Indirect Attributes"] = [
        ", ".join(g)
        for _, g in itertools.groupby(associ_attrs, lambda x: x.rsplit(".", 1)[0])
    ]
    disp["Customized Attributes"] = ", ".join(custom_attrs)
    disp["Birth Command"] = session.symtable[stmt["input"]].birth_statement["command"]
    disp["Associated Datasource"] = session.symtable[stmt["input"]].data_source
    disp["Dependent Variables"] = ", ".join(
        session.symtable[stmt["input"]].dependent_variables
    )

    return None, DisplayDict(disp)


@_debug_logger
@_skip_command_if_empty_input
def disp(stmt, session):
    entity_table = session.symtable[stmt["input"]].entity_table
    transform = stmt.get("transformer")
    if transform and entity_table:
        qry = _transform_query(session.store, entity_table, transform)
    else:
        qry = Query(entity_table)

    qry = _build_query(session.store, entity_table, qry, stmt)
    try:
        cursor = session.store.run_query(qry)
    except InvalidAttr as e:
        var_attr = str(e).split()[-1]
        var_name, _, attr = var_attr.rpartition(".")
        raise MissingEntityAttribute(var_name, attr) from e
    content = cursor.fetchall()

    return None, DisplayDataframe(dedup_ordered_dicts(remove_empty_dicts(content)))


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def get(stmt, session):
    pattern = stmt["stixpattern"]
    local_var_table = stmt["output"] + "_local"
    return_var_table = stmt["output"]
    return_type = stmt["type"]
    limit = stmt.get("limit")
    display = None

    if "variablesource" in stmt:
        input_type = session.symtable[stmt["variablesource"]].type
        output_type = stmt["type"]
        if input_type != output_type:
            raise InvalidECGPattern(
                f"input variable type {input_type} does not match output type {output_type}"
            )
        session.store.filter(
            return_var_table,
            return_type,
            input_type,
            pattern,
        )
        _logger.debug(f"get from variable source \"{stmt['variablesource']}\"")

    elif "datasource" in stmt:
        # rs: RetStruct
        rs = session.data_source_manager.query(
            stmt["datasource"], pattern, session.session_id, session.store, limit
        )
        query_id = rs.load_to_store(session.store)
        session.store.extract(local_var_table, return_type, query_id, pattern)
        local_stage_varstruct = new_var(
            session.store, local_var_table, [], stmt, session.symtable
        )
        _logger.debug(
            f"native GET pattern executed and DB view {local_var_table} extracted."
        )

        # TODO: add a ECGP method to do this directly
        pat_summary = summarize_pattern(pattern)

        local_stage_var_entity_id = get_entity_id_attribute(local_stage_varstruct)
        if (
            pat_summary
            and return_type in pat_summary  # allow extended subgraph
            and len(pat_summary[return_type]) == 1  # only one attr for center node
            and pat_summary[return_type].pop() == local_stage_var_entity_id
        ):
            _logger.debug("To skip prefetch for direct query")
            is_direct_query = True
        else:
            is_direct_query = False

        return_var_table = do_prefetch(
            local_var_table, local_stage_varstruct, session, stmt, not is_direct_query
        )

    else:
        raise KestrelInternalError(f"unknown type of source in {str(stmt)}")

    output = new_var(session.store, return_var_table, [], stmt, session.symtable)

    if not len(output):
        if not return_type.startswith("x-") and return_type not in (
            set(session.store.types()) | set(get_entity_types())
        ):
            display = DisplayWarning(f'unknown entity type "{return_type}"')

    return output, display


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def find(stmt, session):
    # shortcuts
    return_type = stmt["type"]
    input_type = session.symtable[stmt["input"]].type
    local_var_table = stmt["output"] + "_local"

    # init
    rel_query = None
    return_var_table = None

    if return_type in session.store.types():
        input_var_attrs = session.store.columns(input_type)
        return_type_attrs = session.store.columns(return_type)

        # First, get information from local store
        if stmt["relation"] in generic_relations:
            rel_query = compile_generic_relation_to_query(
                return_type, input_type, session.symtable[stmt["input"]].entity_table
            )

        else:
            rel_query = compile_specific_relation_to_query(
                return_type,
                stmt["relation"],
                input_type,
                stmt["reversed"],
                stmt["input"],
                input_var_attrs,
                return_type_attrs,
            )

        # `session.store.assign_query` will generate new entity_table named `local_var_table`
        if rel_query:
            session.store.assign_query(local_var_table, rel_query, return_type)
            local_stage_varstruct = new_var(
                session.store, local_var_table, [], stmt, session.symtable
            )
            return_var_table = do_prefetch(
                local_var_table, local_stage_varstruct, session, stmt
            )

    output = new_var(session.store, return_var_table, [], stmt, session.symtable)
    return output, None


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def join(stmt, session):
    session.store.join(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        stmt["attribute_1"],
        session.symtable[stmt["input_2"]].entity_table,
        stmt["attribute_2"],
    )


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def group(stmt, session):
    if "aggregations" in stmt:
        aggs = [(i["func"], i["attr"], i["alias"]) for i in stmt["aggregations"]]
    else:
        aggs = None
    session.store.group(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        stmt["attributes"],
        aggs,
    )


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def sort(stmt, session):
    session.store.assign(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        op="sort",
        by=stmt["attribute"],
        ascending=stmt["ascending"],
    )


@_debug_logger
@_default_output
@_skip_command_if_empty_input
def apply(stmt, session):
    arg_vars = [session.symtable[v_name] for v_name in stmt["inputs"]]
    display = session.analytics_manager.execute(
        stmt["analytics_uri"], arg_vars, session.session_id, stmt["arguments"]
    )
    return None, display


################################################################
#                       Helper Functions
################################################################


def _set_projection(store, entity_table, query, paths):
    joins, proj = auto_deref(store, entity_table, paths=paths)
    query.joins.extend(joins)
    joined = [j.name for j in query.joins]
    _logger.debug("%s: joining %s", query.table.name, joined)
    if query.proj:
        # Need to merge projections?  More-specific overrides less-specific ("*")
        new_cols = []
        for p in query.proj.cols:
            if not (hasattr(p, "table") and p.table == entity_table and p.name == "*"):
                new_cols.append(p)
        if proj:
            for p in proj.cols:
                if not (
                    hasattr(p, "table") and p.table == entity_table and p.name == "*"
                ):
                    new_cols.append(p)
        query.proj = Projection(new_cols)
    else:
        query.proj = proj


def _get_pred_columns(preds: list):
    for pred in preds:
        if isinstance(pred.lhs, Predicate):
            yield from _get_pred_columns([pred.lhs])
            if isinstance(pred.rhs, Predicate):
                yield from _get_pred_columns([pred.rhs])
        else:
            yield pred.lhs


def _get_filt_columns(filts: list):
    for filt in filts:
        yield from _get_pred_columns(filt.preds)


def _build_query(store, entity_table, qry, stmt, paths=None):
    where = stmt.get("where")
    if where:
        if isinstance(where, Query):
            for j in where.joins:
                _logger.debug("Anchoring JOIN to %s", qry.table.name)
                j.prev_name = qry.table.name
            qry.joins.extend(where.joins)
            for col in _get_filt_columns(where.where):
                if col.table is None:
                    # Need to disambiguate any Predicate columns
                    _logger.debug("Disambiguating predicate for %s", qry.table.name)
                    col.table = qry.table.name
            qry.where.extend(where.where)
        else:
            where.set_table(entity_table)
            qry.append(where)
    attrs = stmt.get("attrs", "*")
    if attrs == "*" and not qry.joins:
        # If user didn't ask for any paths and the where clause didn't
        # result in any joins, fallback to the calling function's list
        # of paths.
        # https://github.com/opencybersecurityalliance/kestrel-lang/issues/312
        cols = paths
    else:
        cols = attrs.split(",")
    _set_projection(store, entity_table, qry, cols)
    sort_by = stmt.get("attribute")
    if sort_by:
        direction = "ASC" if stmt["ascending"] else "DESC"
        qry.append(Order([(sort_by, direction)]))
    limit = stmt.get("limit")
    if limit:
        qry.append(Limit(limit))
    offset = stmt.get("offset")
    if offset:
        qry.append(Offset(offset))
    return qry


def _transform_query(store, entity_table, transform):
    if transform.lower() == "timestamped":
        qry = store.timestamped(entity_table, run=False)
    elif transform.lower() == "addobsid":
        qry = store.extract_observeddata_attribute(
            entity_table, name_of_attribute="id", run=False
        )
    elif transform.lower() == "records":
        qry = store.extract_observeddata_attribute(
            entity_table,
            name_of_attribute=[
                "number_observed",
                "first_observed",
                "last_observed",
                "id",
            ],
            run=False,
        )
    else:
        qry = Query(entity_table)
    return qry



import pytest
import subprocess

from .utils import stixshifter_profile_lab101

@pytest.fixture()
def create_huntflow(tmp_path):
    huntflow = """
procs = GET process FROM stixshifter://lab101
        WHERE name = 'svchost.exe'
        START 2021-01-01T00:00:00Z STOP 2022-01-01T00:00:00Z
"""

    expected_result_lines = ["VARIABLE    TYPE  #(ENTITIES)  #(RECORDS)  directory*  file*  ipv4-addr*  ipv6-addr*  mac-addr*  network-traffic*  process*  user-account*  x-ecs-destination*  x-ecs-network*  x-ecs-process*  x-ecs-source*  x-ecs-user*  x-oca-asset*  x-oca-event*", "   procs process          389        1066        1078   1114        3190        1910       1066              1014       725           1062                2016            2016            2120           2024         2124          1066          2132"]

    huntflow_file = tmp_path / "hunt101.hf"

    with open(huntflow_file, "w") as hf:
        hf.write(huntflow)

    huntflow_file_path = str(huntflow_file.expanduser().resolve())

    return huntflow_file_path, expected_result_lines



def test_cli(create_huntflow, stixshifter_profile_lab101):

    huntflow_file_path, expected_result_lines = create_huntflow
    result = subprocess.run(args = ["kestrel", huntflow_file_path], 
                            universal_newlines = True,
                            stdout = subprocess.PIPE
                           )

    result_lines = result.stdout.splitlines()
    assert result_lines[-3] == expected_result_lines[0]
    assert result_lines[-2] == expected_result_lines[1]


def test_python_module_call(create_huntflow, stixshifter_profile_lab101):

    huntflow_file_path, expected_result_lines = create_huntflow
    result = subprocess.run(args = ["python", "-m", "kestrel", huntflow_file_path], 
                            universal_newlines = True,
                            stdout = subprocess.PIPE
                           )

    result_lines = result.stdout.splitlines()
    assert result_lines[-3] == expected_result_lines[0]
    assert result_lines[-2] == expected_result_lines[1]



import logging
import json
import pandas
import pytest
from multiprocessing import Queue

from kestrel_datasource_stixshifter.connector import check_module_availability
from kestrel_datasource_stixshifter import multiproc
from kestrel_datasource_stixshifter.worker.utils import TransmissionResult
from kestrel_datasource_stixshifter.worker import STOP_SIGN


CONNECTOR_NAME = "elastic_ecs"

SAMPLE_RESULT = TransmissionResult(
    "fake transmission worker",
    True,
    [
        {
            "process": {
                "name": "svchost.exe",
                "pid": 1744,
                "entity_id": "{8dfc401c-f5a9-6068-2300-000000001300}",
                "executable": "C:\\Windows\\System32\\svchost.exe",
            },
            "@timestamp": "2021-04-10T05:35:45.333Z",
            "destination": {
                "port": 49343,
                "ip": "127.0.0.1",
                "domain": "thl-win.example.com",
            },
            "host": {
                "hostname": "thl-win",
                "os": {
                    "build": "17763.1817",
                    "kernel": "10.0.17763.1817 (WinBuild.160101.0800)",
                    "name": "Windows Server 2019 Standard",
                    "family": "windows",
                    "version": "10.0",
                    "platform": "windows",
                },
                "ip": ["2001:db8:a::123", "198.51.100.101", "10.0.0.141"],
                "name": "thl-win.example.com",
                "id": "d3f820f9-f601-4dd8-bd48-5c13549a6027",
                "mac": ["00:25:96:12:34:56"],
                "architecture": "x86_64",
            },
            "source": {
                "port": 49343,
                "ip": "127.0.0.1",
                "domain": "thl-win.example.com",
            },
            "event": {
                "code": 3,
                "provider": "Microsoft-Windows-Sysmon",
                "kind": "event",
                "created": "2021-04-10T05:35:43.069Z",
                "module": "sysmon",
                "action": "Network connection detected (rule: NetworkConnect)",
                "category": ["network"],
                "type": ["connection", "start", "protocol"],
            },
            "user": {"domain": "NT AUTHORITY", "name": "SYSTEM"},
            "network": {
                "community_id": "1:ZXhhbXBsZQ==",
                "protocol": "-",
                "transport": "udp",
                "type": "ipv4",
                "direction": "egress",
            },
            "tags": ["beats_input_codec_plain_applied"],
        }
    ],
    100,
    None,
)


def test_stixshifter_translate():
    query_id = "8df266aa-2901-4a94-ace9-a4403e310fa1"
    check_module_availability(CONNECTOR_NAME)

    input_queue = Queue()
    output_queue = Queue()

    with multiproc.translate(
        CONNECTOR_NAME,
        {"id": "identity--" + query_id, "name": CONNECTOR_NAME},
        {},
        None,
        False,
        input_queue,
        output_queue,
        1,
    ) as translators:

        input_queue.put(SAMPLE_RESULT)
        input_queue.put(STOP_SIGN)

        for result in multiproc.read_translated_results(output_queue, 1):
            id_object = result["objects"][0]
            assert id_object["id"] == "identity--" + query_id
            assert id_object["name"] == CONNECTOR_NAME

    for translator in translators:
        assert translator.is_alive() == False


def test_stixshifter_translate_with_bundle_writing_to_disk(tmpdir):
    query_id = "8df266aa-2901-4a94-ace9-a4403e310fa1"
    check_module_availability(CONNECTOR_NAME)
    cache_bundle_path_prefix = str(tmpdir.join("test"))
    offset_str = str(SAMPLE_RESULT.offset).zfill(32)
    cache_bundle_path = cache_bundle_path_prefix + f"_{offset_str}.json"

    input_queue = Queue()
    output_queue = Queue()

    with multiproc.translate(
        CONNECTOR_NAME,
        {"id": "identity--" + query_id, "name": CONNECTOR_NAME},
        {},
        cache_bundle_path_prefix,
        False,
        input_queue,
        output_queue,
        1,
    ) as translators:

        input_queue.put(SAMPLE_RESULT)
        input_queue.put(STOP_SIGN)

        for result in multiproc.read_translated_results(output_queue, 1):
            pass

        with open(cache_bundle_path, "r") as bundle_fp:
            bundle = json.load(bundle_fp)
            id_object = bundle["objects"][0]
            assert id_object["id"] == "identity--" + query_id
            assert id_object["name"] == CONNECTOR_NAME

    for translator in translators:
        assert translator.is_alive() == False


def test_fast_translate():
    query_id = "8df266aa-2901-4a94-ace9-a4403e310fa1"
    check_module_availability(CONNECTOR_NAME)

    input_queue = Queue()
    output_queue = Queue()

    with multiproc.translate(
        CONNECTOR_NAME,
        {"id": "identity--" + query_id, "name": CONNECTOR_NAME},
        {},
        None,
        True,
        input_queue,
        output_queue,
        1,
    ) as translators:

        input_queue.put(SAMPLE_RESULT)
        input_queue.put(STOP_SIGN)

        for result in multiproc.read_translated_results(output_queue, 1):
            assert isinstance(result, pandas.DataFrame)
            assert result.empty == False

    for translator in translators:
        assert translator.is_alive() == False


def test_stixshifter_fast_translate_with_parquet_writing_to_disk(tmpdir):
    query_id = "8df266aa-2901-4a94-ace9-a4403e310fa1"
    check_module_availability(CONNECTOR_NAME)
    cache_parquet_path_prefix = str(tmpdir.join("test"))
    offset_str = str(SAMPLE_RESULT.offset).zfill(32)
    cache_parquet_path = cache_parquet_path_prefix + f"_{offset_str}.parquet"

    input_queue = Queue()
    output_queue = Queue()

    with multiproc.translate(
        CONNECTOR_NAME,
        {"id": "identity--" + query_id, "name": CONNECTOR_NAME},
        {},
        cache_parquet_path_prefix,
        True,
        input_queue,
        output_queue,
        1,
    ) as translators:

        input_queue.put(SAMPLE_RESULT)
        input_queue.put(STOP_SIGN)

        for result in multiproc.read_translated_results(output_queue, 1):
            pass

        df = pandas.read_parquet(cache_parquet_path)

    for translator in translators:
        assert translator.is_alive() == False



import pytest
import json

from kestrel.session import Session


def test_display_block_summary_to_json():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        d = s.execute(stmt)
    correct_json = json.loads(
        '{"display": "execution summary", "data": {"variables updated": [{"VARIABLE": "newvar", "TYPE": "process", "#(ENTITIES)": 2, "#(RECORDS)": 2, "process*": 0}], "footnotes": ["*Number of related records cached."]}}'
    )
    output_json = json.loads(d[0].to_json())
    del output_json["data"]["execution time"]
    assert output_json == correct_json


def test_display_block_summary_to_dict():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        d = s.execute(stmt)
    correct_dict = {
        "display": "execution summary",
        "data": {
            "variables updated": [
                {
                    "VARIABLE": "newvar",
                    "TYPE": "process",
                    "#(ENTITIES)": 2,
                    "#(RECORDS)": 2,
                    "process*": 0,
                }
            ],
            "footnotes": ["*Number of related records cached."],
        },
    }
    output_dict = d[0].to_dict()
    del output_dict["data"]["execution time"]
    assert output_dict == correct_dict


def test_display_block_summary_get_from_variable():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "97"}
             , {"type": "process", "name": "explorer.exe", "pid": "98"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
exp = GET process from newvar WHERE [process:name = 'explorer.exe']
"""
        d = s.execute(stmt)
    correct_dict = {
        "display": "execution summary",
        "data": {
            "variables updated": [
                {
                    "VARIABLE": "newvar",
                    "TYPE": "process",
                    "#(ENTITIES)": 4,
                    "#(RECORDS)": 4,
                    "process*": 0,
                },
                {
                    "VARIABLE": "exp",
                    "TYPE": "process",
                    "#(ENTITIES)": 3,
                    "#(RECORDS)": 3,
                    "process*": 0,
                },
            ],
            "footnotes": ["*Number of related records cached."],
        },
    }
    output_dict = d[0].to_dict()
    del output_dict["data"]["execution time"]
    assert output_dict == correct_dict



import pytest
import subprocess

from kestrel_datasource_stixshifter.diagnosis import Diagnosis
from .utils import stixshifter_profile_lab101, stixshifter_profile_ecs


def test_diagnosis(stixshifter_profile_lab101):
    pattern = " ".join(
        [
            "[ipv4-addr:value != '255.255.255.255']",
            "START t'2000-01-01T00:00:00.000Z' STOP t'3000-01-01T00:00:00.000Z'",
        ]
    )
    diag = Diagnosis("lab101")
    diag.diagnose_config()
    diag.diagnose_ping()
    assert pattern == diag.diagnose_translate_query(pattern)["queries"][0]
    res = diag.diagnose_run_query_and_retrieval_result([pattern], 1)
    assert len(res) == 1 and res[0] == 533


def test_cli(stixshifter_profile_lab101):
    expected_output = """
## Diagnose: config verification

#### Kestrel specific config
retrieval batch size: 2000
cool down after transmission: 0
enable fast translation: False

#### Config to be passed to stix-shifter
connector name: stix_bundle
connection object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#connection]:
{
    "host": "https://github.com/opencybersecurityalliance/data-bucket-kestrel/blob/main/stix-bundles/lab101.json?raw=true",
    "options": {
        "result_limit": 4000,
        "timeout": 60
    }
}
configuration object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#configuration]:
{
    "auth": {
        "username": null,
        "password": null
    }
}

## Diagnose: stix-shifter query translation

#### Input: STIX pattern
[ipv4-addr:value != '255.255.255.255'] START t'2000-01-01T00:00:00.000Z' STOP t'3000-01-01T00:00:00.000Z'

#### Output: 1 data source native queries
[ipv4-addr:value != '255.255.255.255'] START t'2000-01-01T00:00:00.000Z' STOP t'3000-01-01T00:00:00.000Z'

## Diagnose: stix-shifter to data source connection (network, auth)

#### Results from stixshifter transmission.ping()
{
    "success": true
}

## Diagnose: stix-shifter query execution: <=1 batch(s)

#### data retrieval results:
one batch retrieved: 533 entries

## Diagnose: stix-shifter query execution: <=5 batch(s)

#### data retrieval results:
one batch retrieved: 533 entries
"""

    result = subprocess.run(
        args=["stix-shifter-diag", "lab101"],
        universal_newlines=True,
        stdout=subprocess.PIPE,
    )

    result_lines = result.stdout.splitlines()
    result_lines = [x for x in result_lines if x]
    expected_lines = expected_output.splitlines()
    expected_lines = [x for x in expected_lines if x]
    for x, y in zip(result_lines, expected_lines):
        assert x == y


def test_cli_ecs(stixshifter_profile_ecs):
    expected_output = """
## Diagnose: config verification

#### Kestrel specific config
retrieval batch size: 2000
cool down after transmission: 0
enable fast translation: False

#### Config to be passed to stix-shifter
connector name: elastic_ecs
connection object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#connection]:
{
    "host": "elastic.securitylog.company.com",
    "port": 9200,
    "selfSignedCert": false,
    "indices": "host101",
    "options": {
        "result_limit": 4000,
        "timeout": 60
    }
}
configuration object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#configuration]:
{
    "auth": {
        "id": "********",
        "api_key": "********"
    }
}

## Diagnose: stix-shifter query translation

#### Input: STIX pattern
[x-oca-asset:device_id = '123456'] START t'2000-01-01T00:00:00.000Z' STOP t'3000-01-01T00:00:00.000Z'

#### Output: 1 data source native queries
(host.id : "123456" OR observer.serial_number : "123456") AND (@timestamp:["2000-01-01T00:00:00.000Z" TO "3000-01-01T00:00:00.000Z"])
"""

    result = subprocess.run(
        args=[
            "stix-shifter-diag",
            "-p",
            "[x-oca-asset:device_id = '123456'] START t'2000-01-01T00:00:00.000Z' STOP t'3000-01-01T00:00:00.000Z'",
            "-t",
            "ecs",
        ],
        universal_newlines=True,
        stdout=subprocess.PIPE,
    )

    result_lines = result.stdout.splitlines()
    result_lines = [x for x in result_lines if x]
    expected_lines = expected_output.splitlines()
    expected_lines = [x for x in expected_lines if x]
    for x, y in zip(result_lines, expected_lines):
        assert x == y
